#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if

from .version import __version__, VERSION
from .availability import AVAILABLE_BACKENDS

from .pure_python import RGB, Luma, Sums

try:
    from .compiled_cython import rgb_sums, sums_to_luma
except ImportError:
    from .pure_python import rgb_sums, sums_to_luma
# except

try:
    from .native_ctypes import jpeg_intensities, png_intensities, rgb_luma_from_filename
except ImportError:
    try:
        from .native_cffi import jpeg_intensities, png_intensities, rgb_luma_from_filename
    except ImportError:
        try:
            from .compiled_cffi import jpeg_intensities, png_intensities, rgb_luma_from_filename
        except ImportError:
            try:
                from .compiled_cython import rgb_luma_from_filename
                jpeg_intensities = rgb_luma_from_filename
                png_intensities = rgb_luma_from_filename
            except ImportError:
                from .pure_python import rgb_luma_from_filename
                jpeg_intensities = rgb_luma_from_filename
                png_intensities = rgb_luma_from_filename
            # end try
        # end try
    # end try
# end try






from mimetypes import guess_type
from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    import cffi
    import ctypes
# end if

from .pure_python import Luma


def convert_struct_to_luma(struct) -> Luma:
    return Luma(nw=struct.nw, ne=struct.ne, sw=struct.sw, se=struct.se)
# end def


def lib_jpeg_intensities(filename, lib: Union['cffi.FFI', 'ctypes.CDLL']) -> Luma:
    # noinspection PyUnresolvedReferences
    return convert_struct_to_luma(lib.jpeg_intensities(filename))
# end def


def lib_png_intensities(filename, lib: Union['cffi.FFI', 'ctypes.CDLL']) -> Luma:
    # noinspection PyUnresolvedReferences
    return convert_struct_to_luma(lib.png_intensities(filename))
# end def


def lib_rgb_luma_from_filename(filename: str, argument_filename, lib):
    (mime_type, encoding) = guess_type(filename)
    if mime_type == 'image/png':
        return lib_png_intensities(argument_filename, lib)
    elif mime_type == 'image/jpeg':
        return lib_jpeg_intensities(argument_filename, lib)
    else:
        raise ValueError('Unknown mime.')
    # end if
# end def



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if


import ctypes
import pathlib
from luckydonaldUtils.encoding import to_binary as b

from .pure_python import Luma
from .shared_cffi import lib_jpeg_intensities, lib_png_intensities, lib_rgb_luma_from_filename

# Load the shared library into ctypes
libname = pathlib.Path().absolute() / "_native_code" / "lib" / "libimage_intensities.so"
if not libname.exists():
    raise ImportError(f'Could not find library at {libname!s}.')
# end if


libname = str(libname)
logger.debug(f'Image intensities library path: {libname!r}')

c_lib = ctypes.CDLL(libname)


class IntensityData(ctypes.Structure):
    _fields_ = [
        ("nw", ctypes.c_double),
        ("ne", ctypes.c_double),
        ("sw", ctypes.c_double),
        ("se", ctypes.c_double),
    ]
# end class


c_lib.jpeg_intensities.restype = IntensityData
c_lib.png_intensities.restype = IntensityData

# struct intensity_data jpeg_intensities(const char *file_name);
# struct intensity_data png_intensities(const char *file_name);


def jpeg_intensities(file_name: str) -> Luma:
    return lib_jpeg_intensities(ctypes.c_char_p(b(file_name)), c_lib)
# end def


def png_intensities(file_name: str) -> Luma:
    return lib_png_intensities(ctypes.c_char_p(b(file_name)), c_lib)
# end def


def rgb_luma_from_filename(file_name: str) -> Luma:
    return lib_rgb_luma_from_filename(file_name, ctypes.c_char_p(b(file_name)), c_lib)
# end def



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from luckydonaldUtils.logger import logging

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if

AVAILABLE_BACKENDS = {
    "pure_python": True,
    "compiled_cython": None,
    "compiled_cffi": None,
    "native_cffi": None,
    "native_ctypes": None,
}

try:
    from image_intensities.pure_python import rgb_sums, sums_to_luma, rgb_luma_from_filename
    AVAILABLE_BACKENDS["pure_python"] = True
except Exception as e:
    AVAILABLE_BACKENDS["pure_python"] = False
    print(f'pure_python failed: {e!s}')
# end try

try:
    from image_intensities.compiled_cython import rgb_sums, sums_to_luma, rgb_luma_from_filename
    AVAILABLE_BACKENDS["compiled_cython"] = True
except Exception as e:
    AVAILABLE_BACKENDS["compiled_cython"] = False
    print(f'compiled_cython failed: {e!s}')
# end try

try:
    from image_intensities.compiled_cffi import jpeg_intensities, png_intensities, rgb_luma_from_filename
    AVAILABLE_BACKENDS["compiled_cffi"] = True
except Exception as e:
    AVAILABLE_BACKENDS["compiled_cffi"] = False
    print(f'compiled_cffi failed: {e!s}')
# end try

try:
    from image_intensities.native_cffi import jpeg_intensities, png_intensities, rgb_luma_from_filename
    AVAILABLE_BACKENDS["native_cffi"] = True
except Exception as e:
    AVAILABLE_BACKENDS["native_cffi"] = False
    print(f'native_cffi failed: {e!s}')
# end try

try:
    from image_intensities.native_ctypes import jpeg_intensities, png_intensities, rgb_luma_from_filename
    AVAILABLE_BACKENDS["native_ctypes"] = True
except Exception as e:
    AVAILABLE_BACKENDS["native_ctypes"] = False
    print(f'native_ctypes failed: {e!s}')
# end try




"""
The “out-of-line”, “API mode” gives you the most flexibility and speed to access a C library at the level of C, instead of at the binary level:
"""
# in a separate file "package/foo_build.py"

import cffi
from pathlib import Path

with open(str(Path(__file__).absolute().parent / 'build_cffi.c-files.txt'), 'r') as f:
    sources = f.read().strip().splitlines()
# end with

ffibuilder = cffi.FFI()
ffibuilder.set_source(
    module_name="image_intensities._native_code._image_intensities",
    source="""
    #include "_native_code/definitions.h"
    #include "_native_code/turbojpeg.h"
    """,
    include_dirs=[  # -I
        "./_native_code/turbojpeg",
        "./_native_code"
    ],
    libraries=[  # -L
        "./_native_code/jpeg",
        "./_native_code/png",
    ],
    sources=[f'./_native_code/{filename}' for filename in sources if filename],
    extra_compile_args=[
        "-std=c99",
        "-fPIC",
        "-O3",
        "-DPPM_SUPPORTED",
        "-DBMP_SUPPORTED",
    ],
)
ffibuilder.cdef("""
    struct intensity_data {
        double nw;
        double ne;
        double sw;
        double se;
        int error;
    };

    struct intensity_data jpeg_intensities(const char *file_name);
    struct intensity_data png_intensities(const char *file_name);
""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)
# end if



from luckydonaldUtils.encoding import to_binary as b
from .pure_python import Luma
from .shared_cffi import lib_jpeg_intensities, lib_png_intensities, lib_rgb_luma_from_filename


# noinspection PyProtectedMember,PyUnresolvedReferences
from ._native_code._image_intensities import ffi, lib


def jpeg_intensities(filename) -> Luma:
    return lib_jpeg_intensities(b(filename), lib)
# end def


def png_intensities(filename) -> Luma:
    return lib_png_intensities(b(filename), lib)
# end def


def rgb_luma_from_filename(filename) -> Luma:
    return lib_rgb_luma_from_filename(filename, b(filename), lib)
# end def



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from cffi import FFI
from luckydonaldUtils.logger import logging
from luckydonaldUtils.encoding import to_binary as b

__author__ = 'luckydonald'

logger = logging.getLogger(__name__)
if __name__ == '__main__':
    logging.add_colored_handler(level=logging.DEBUG)
# end if

from pathlib import Path
from .pure_python import Luma
from .shared_cffi import lib_jpeg_intensities, lib_rgb_luma_from_filename
from .shared_cffi import lib_png_intensities

so_files = Path(__file__).joinpath('_native_code').absolute().glob("**/*image_intensities*.so")
try:
    so_file = str(next(so_files))
except:
    logger.warning('Loading optimized library via CFFI failed, file not found.')
    raise ImportError('File not found.')
# end if
ffi = FFI()
ffi.cdef("""
    extern struct intensity_data {
        double nw;
        double ne;
        double sw;
        double se;
        int error;
    } intensity_data;

    struct intensity_data jpeg_intensities(const char *file_name);
    struct intensity_data png_intensities(const char *file_name);
""")
try:
    lib = ffi.dlopen(so_file)
except OSError as e:
    logger.warning('Loading optimized library via CFFI failed:', exc_info=True)
    raise ImportError('OSError: ' + str(e))
# end if


def jpeg_intensities(filename) -> Luma:
    return lib_jpeg_intensities(b(filename), lib)
# end def


def png_intensities(filename) -> Luma:
    return lib_png_intensities(b(filename), lib)
# end def


def rgb_luma_from_filename(filename) -> Luma:
    return lib_rgb_luma_from_filename(filename, b(filename), lib)
# end def
